@@ -19,6 +19,7 @@
 #include "Gui/ExportMenu.h"
 #include "Gui/HelpMenu.h"
 #include "Gui/OptionsMenu.h"
+#include "Gui/InlineMessageModel.h"
 #include "Platforms/ImagePlatformXcb.h"
 #include "Platforms/VideoPlatform.h"
 #include "ShortcutActions.h"
@@ -128,6 +129,7 @@ SpectacleCore::SpectacleCore(QObject *parent)
     connect(SelectionEditor::instance(), &SelectionEditor::accepted,
             this, [this](const QRectF &rect, const ExportManager::Actions &actions){
         ExportManager::instance()->updateTimestamp();
+        m_returnToViewer = m_startMode == StartMode::Gui;
         if (m_videoMode) {
             const auto captureWindows = CaptureWindow::instances();
             SpectacleWindow::setVisibilityForAll(QWindow::Hidden);
@@ -160,19 +162,25 @@ SpectacleCore::SpectacleCore(QObject *parent)
             deleteWindows();
             m_annotationDocument->cropCanvas(rect);
             syncExportImage();
+            const auto &exportActions = actions & ExportManager::AnyAction ? actions : autoExportActions();
+            const bool willQuit = exportActions.testFlag(ExportManager::AnyAction) //
+                && exportActions.testFlag(ExportManager::UserAction) //
+                && Settings::quitAfterSaveCopyExport();
+            m_returnToViewer &= !willQuit;
             showViewerIfGuiMode();
             SpectacleWindow::setTitleForAll(SpectacleWindow::Unsaved);
             ExportManager::instance()->scanQRCode();
-            const auto &exportActions = actions & ExportManager::AnyAction ? actions : autoExportActions();
             ExportManager::instance()->exportImage(exportActions, outputUrl());
         }
     });
 
     connect(imagePlatform, &ImagePlatform::newScreenshotTaken, this, [this](const QImage &image){
+        InlineMessageModel::instance()->clear();
         m_annotationDocument->clearAnnotations();
         m_annotationDocument->setBaseImage(image);
         setExportImage(image);
         ExportManager::instance()->updateTimestamp();
+        m_returnToViewer = true;
         showViewerIfGuiMode();
         SpectacleWindow::setTitleForAll(SpectacleWindow::Unsaved);
         ExportManager::instance()->scanQRCode();
@@ -180,6 +188,7 @@ SpectacleCore::SpectacleCore(QObject *parent)
         setVideoMode(false);
     });
     connect(imagePlatform, &ImagePlatform::newCroppableScreenshotTaken, this, [this](const QImage &image) {
+        InlineMessageModel::instance()->clear();
         setVideoMode(false);
         m_annotationDocument->clearAnnotations();
         m_annotationDocument->setBaseImage(image);
@@ -192,8 +201,7 @@ SpectacleCore::SpectacleCore(QObject *parent)
     // The behavior happens to be very similar right now, so we use a shared base function
     auto onScreenshotOrRecordingFailed = [this](const QString &message,
                                                 QString uiMessage,
-                                                void (SpectacleCore::*dBusFailedSignal)(const QString &),
-                                                void (ViewerWindow::*showViewerMessage)(const QString &)) {
+                                                void (SpectacleCore::*dBusFailedSignal)(const QString &)) {
         if (!message.isEmpty()) {
             uiMessage = uiMessage % u"\n"_s % message;
         }
@@ -219,28 +227,46 @@ SpectacleCore::SpectacleCore(QObject *parent)
             return;
         }
         case StartMode::Gui: {
-            if (!ViewerWindow::instance() && hasVisibleWindow) {
-                showErrorMessage(uiMessage);
-                return;
-            } else if (!ViewerWindow::instance()) {
-                initViewerWindow(ViewerWindow::Dialog);
+            if (!ViewerWindow::instance()) {
+                InlineMessageModel::instance()->clear();
+                if (hasVisibleWindow) {
+                    showErrorMessage(uiMessage);
+                    return;
+                } else {
+                    initViewerWindow(ViewerWindow::Dialog);
+                }
             }
             qWarning().noquote() << message;
             auto viewer = ViewerWindow::instance();
             viewer->setVisible(true);
-            (viewer->*showViewerMessage)(uiMessage);
+            InlineMessageModel::instance()->push(InlineMessageModel::Error, uiMessage);
             return;
         }
         }
     };
     connect(imagePlatform, &ImagePlatform::newScreenshotFailed, this, [onScreenshotOrRecordingFailed](const QString message) {
         auto uiMessage = i18nc("@info", "An error occurred while taking a screenshot.");
-        onScreenshotOrRecordingFailed(message, uiMessage, &SpectacleCore::dbusScreenshotFailed, &ViewerWindow::showScreenshotFailedMessage);
+        onScreenshotOrRecordingFailed(message, uiMessage, &SpectacleCore::dbusScreenshotFailed);
+    });
+    connect(imagePlatform, &ImagePlatform::newScreenshotCanceled, this, [this]() {
+        if (m_startMode != StartMode::Gui || !m_returnToViewer || isGuiNull()) {
+            Q_EMIT allDone();
+            return;
+        }
+        SpectacleWindow::setTitleForAll(SpectacleWindow::Previous);
+        const auto windows = SpectacleWindow::instances();
+        if (windows.empty()) {
+            initViewerWindow(ViewerWindow::Viewer);
+            return;
+        }
+        for (auto w : windows) {
+            w->setVisible(true);
+        }
     });
 
     auto videoPlatform = m_videoPlatform.get();
-    connect(videoPlatform, &VideoPlatform::recordingChanged, this, [this](bool isRecording) {
-        if (isRecording) {
+    connect(videoPlatform, &VideoPlatform::recordingStateChanged, this, [this](VideoPlatform::RecordingState state) {
+        if (state == VideoPlatform::RecordingState::Recording) {
             static const auto recordingIcon = u":/icons/256-status-media-recording.webp"_s;
             static const auto recordingStartedIcon = u":/icons/256-status-media-recording-started.webp"_s;
             static const auto recordingPulseIcon = u":/icons/256-status-media-recording-pulse.webp"_s;
@@ -255,7 +281,21 @@ SpectacleCore::SpectacleCore(QObject *parent)
             });
             const auto messageTitle = i18nc("recording notification title", "Spectacle is Recording");
             const auto messageBody = i18nc("recording notification message", "Click the system tray icon to finish recording");
-            s_systemTrayIcon->showMessage(messageTitle, messageBody, u"media-record"_s, 4000);
+            auto notification = new KNotification(u"notification"_s, KNotification::CloseOnTimeout | KNotification::DefaultEvent, this);
+            notification->setTitle(messageTitle);
+            notification->setText(messageBody);
+            notification->setIconName(u"media-record"_s);
+            // Whether the notification is transient and should not be kept in history.
+            notification->setHint(u"transient"_s, true);
+            // Can't set notification duration with KNotification directly.
+            // Also see https://bugs.kde.org/show_bug.cgi?id=503838
+            QTimer::singleShot(4000, notification, &KNotification::close);
+            connect(m_videoPlatform.get(), &VideoPlatform::recordingStateChanged, notification, [notification](VideoPlatform::RecordingState state) {
+                if (state != VideoPlatform::RecordingState::Recording) {
+                    notification->close();
+                }
+            });
+            notification->sendEvent();
             if (!QMovie::supportedFormats().contains("webp"_ba)) {
                 const auto messageTitle = i18nc("missing webp support notification title", "WebP support is missing.");
                 const auto messageBody = i18nc("missing webp support notification message", "Please install Qt Image Formats to get animated system tray icons for Spectacle, and then report this packaging issue to your distributor.");
@@ -293,14 +333,35 @@ SpectacleCore::SpectacleCore(QObject *parent)
                 initPulseAnimation();
             });
             startedAnimation->start();
+        } else if (state == VideoPlatform::RecordingState::Rendering && s_systemTrayIcon) {
+            const auto messageTitle = i18nc("recording notification title", "Spectacle is Finishing the Recording");
+            const auto messageBody = i18nc("recording notification message", "Please wait");
+            s_systemTrayIcon->setToolTipTitle(i18nc("@info:tooltip title for rendering tray icon", //
+                                                    "Spectacle is Finishing the Recording"));
+            auto subtitle = i18nc("@info:tooltip subtitle for rendering tray icon", //
+                                  "Time recorded: %1\n" //
+                                  "Click to stop rendering early (this will lose data)",
+                                  recordedTime());
+            auto notification = new KNotification(u"notification"_s, KNotification::CloseOnTimeout | KNotification::DefaultEvent, this);
+            notification->setTitle(messageTitle);
+            notification->setText(messageBody);
+            notification->setIconName(u"process-working-symbolic"_s);
+            notification->setHint(u"transient"_s, true);
+            connect(m_videoPlatform.get(), &VideoPlatform::recordingStateChanged, notification, [notification](VideoPlatform::RecordingState state) {
+                if (state != VideoPlatform::RecordingState::Rendering) {
+                    notification->close();
+                }
+            });
+            notification->sendEvent();
+            s_systemTrayIcon->setToolTipSubTitle(subtitle);
         } else {
             s_systemTrayIcon.reset();
             m_captureWindows.clear();
         }
     });
-    connect(videoPlatform, &VideoPlatform::recordedTimeChanged, this, [this] {
+    connect(videoPlatform, &VideoPlatform::recordedTimeChanged, this, [this, videoPlatform] {
         Q_EMIT recordedTimeChanged();
-        if (!s_systemTrayIcon) {
+        if (!s_systemTrayIcon || videoPlatform->recordingState() != VideoPlatform::RecordingState::Recording) {
             return;
         }
         auto subtitle = i18nc("@info:tooltip subtitle for recording tray icon", //
@@ -314,15 +375,23 @@ SpectacleCore::SpectacleCore(QObject *parent)
         ExportManager::instance()->exportVideo(autoExportActions() | ExportManager::Save, fileUrl, videoOutputUrl());
     });
     connect(videoPlatform, &VideoPlatform::recordingCanceled, this, [this] {
-        if (m_startMode != StartMode::Gui || isGuiNull()) {
+        if (m_startMode != StartMode::Gui || !m_returnToViewer || isGuiNull()) {
             Q_EMIT allDone();
             return;
         }
         SpectacleWindow::setTitleForAll(SpectacleWindow::Previous);
+        const auto windows = SpectacleWindow::instances();
+        if (windows.empty()) {
+            initViewerWindow(ViewerWindow::Viewer);
+            return;
+        }
+        for (auto w : windows) {
+            w->setVisible(true);
+        }
     });
     connect(videoPlatform, &VideoPlatform::recordingFailed, this, [onScreenshotOrRecordingFailed](const QString &message){
         auto uiMessage = i18nc("@info", "An error occurred while attempting to record the screen.");
-        onScreenshotOrRecordingFailed(message, uiMessage, &SpectacleCore::dbusRecordingFailed, &ViewerWindow::showRecordingFailedMessage);
+        onScreenshotOrRecordingFailed(message, uiMessage, &SpectacleCore::dbusRecordingFailed);
     });
     connect(videoPlatform, &VideoPlatform::regionRequested, this, [this] {
         SelectionEditor::instance()->reset();
@@ -363,14 +432,18 @@ SpectacleCore::SpectacleCore(QObject *parent)
         if (actions & ExportManager::AnySave) {
             SpectacleWindow::setTitleForAll(SpectacleWindow::Saved, url.fileName());
             if (actions & ExportManager::CopyImage) {
-                viewerWindow->showSavedAndCopiedMessage(url);
+                auto text = xi18nc("@info", "The screenshot was copied to the clipboard and saved as <link url=\"%1\">%2</link>", url.toString().toHtmlEscaped(), url.fileName().toHtmlEscaped());
+                InlineMessageModel::instance()->push(InlineMessageModel::Saved, text, url);
             } else if (actions & ExportManager::CopyPath) {
-                viewerWindow->showSavedAndLocationCopiedMessage(url);
+                auto text = xi18nc("@info", "The screenshot has been saved as <link url=\"%1\">%2</link> and its location has been copied to clipboard", url.toString().toHtmlEscaped(), url.fileName().toHtmlEscaped());
+                InlineMessageModel::instance()->push(InlineMessageModel::Saved, text, url);
             } else {
-                viewerWindow->showSavedMessage(url);
+                auto text = xi18nc("@info", "The screenshot was saved as <link url=\"%1\">%2</link>", url.toString().toHtmlEscaped(), url.fileName().toHtmlEscaped());
+                InlineMessageModel::instance()->push(InlineMessageModel::Saved, text, url);
             }
         } else if (actions & ExportManager::CopyImage) {
-            viewerWindow->showCopiedMessage();
+            auto text = i18nc("@info", "The screenshot has been copied to the clipboard.");
+            InlineMessageModel::instance()->push(InlineMessageModel::Copied, text);
         }
     };
     connect(exportManager, &ExportManager::imageExported, this, onImageExported);
@@ -400,12 +473,15 @@ SpectacleCore::SpectacleCore(QObject *parent)
         if (actions & ExportManager::AnySave) {
             SpectacleWindow::setTitleForAll(SpectacleWindow::Saved, url.fileName());
             if (actions & ExportManager::CopyPath) {
-                viewerWindow->showSavedAndLocationCopiedMessage(url, true);
+                auto text = xi18nc("@info", "The video has been saved as <link url=\"%1\">%2</link> and its location has been copied to clipboard", url.toString().toHtmlEscaped(), url.fileName().toHtmlEscaped());
+                InlineMessageModel::instance()->push(InlineMessageModel::Saved, text, url);
             } else {
-                viewerWindow->showSavedMessage(url, true);
+                auto text = xi18nc("@info", "The video was saved as <link url=\"%1\">%2</link>", url.toString().toHtmlEscaped(), url.fileName().toHtmlEscaped());
+                InlineMessageModel::instance()->push(InlineMessageModel::Saved, text, url);
             }
         } else if (actions & ExportManager::CopyPath) {
-            viewerWindow->showLocationCopiedMessage();
+            auto text = i18nc("@info", "The video location has been copied to the clipboard.");
+            InlineMessageModel::instance()->push(InlineMessageModel::Copied, text);
         }
     };
     connect(exportManager, &ExportManager::videoExported, this, onVideoExported);
@@ -415,7 +491,22 @@ SpectacleCore::SpectacleCore(QObject *parent)
         if (!viewerWindow) {
             return;
         }
-        viewerWindow->showQRCodeScannedMessage(result);
+        // Thanks to: https://stackoverflow.com/questions/1500260/detect-urls-in-text-with-javascript
+        using QRE = QRegularExpression;
+        static const QRE urlRegex(u"(\\b(https?|ftp|file)://[-A-Z0-9+&@#/%?=~_|!:,.;]*[-A-Z0-9+&@#/%=~_|])"_s, QRE::CaseInsensitiveOption);
+        static const auto linkifier = u"<a href=\"\\1\">\\1</a>"_s;
+        auto text = [](const QVariant &result) -> QString {
+            if (result.typeId() == QMetaType::QString) {
+                // Not using xi18nc because it doesn't work with the link replacement logic.
+                // Also see https://invent.kde.org/graphics/spectacle/-/merge_requests/432#note_1111125
+                auto text = i18nc("@info", //
+                                  "QR Code found: %1",
+                                  result.toString().toHtmlEscaped().replace(urlRegex, linkifier));
+                return u"<html>" % text % u"</html>";
+            }
+            return i18nc("@info", "Found QR code with binary content.");
+        }(result);
+        InlineMessageModel::instance()->push(InlineMessageModel::Scanned, text, result);
     };
     connect(exportManager, &ExportManager::qrCodeScanned, this, onQRCodeScanned);
 
@@ -578,7 +669,7 @@ void SpectacleCore::activate(const QStringList &arguments, const QString &workin
         }
     }
 
-    if (parser.optionNames().size() > 0 || m_startMode != StartMode::Gui) {
+    if (parser.optionNames().size() > 0 || m_startMode != StartMode::Gui || !m_returnToViewer) {
         // Delete windows if we have CLI options or not in GUI mode.
         // We don't want to delete them otherwise because that will mess with the
         // settings for PrintScreen key behavior.
@@ -643,6 +734,7 @@ void SpectacleCore::activate(const QStringList &arguments, const QString &workin
         // QFileInfo::exists() only works with local files.
         auto existingLocalFile = m_editExistingUrl.toLocalFile();
         if (QFileInfo::exists(existingLocalFile)) {
+            InlineMessageModel::instance()->clear();
             // If editing an existing image, open the annotation editor.
             // This QImage constructor only works with local files or Qt resource file names.
             QImage existingImage(existingLocalFile);
@@ -673,20 +765,33 @@ void SpectacleCore::activate(const QStringList &arguments, const QString &workin
     // Determine grab mode
     using CaptureMode = CaptureModeModel::CaptureMode;
     using GrabMode = ImagePlatform::GrabMode;
-    GrabMode grabMode = GrabMode::AllScreens; // Default to all screens
-    if (m_cliOptions[Option::Fullscreen]) {
-        grabMode = GrabMode::AllScreens;
-    } else if (m_cliOptions[Option::Current]) {
-        grabMode = GrabMode::CurrentScreen;
-    } else if (m_cliOptions[Option::ActiveWindow]) {
-        grabMode = GrabMode::ActiveWindow;
-    } else if (m_cliOptions[Option::Region]) {
-        grabMode = GrabMode::PerScreenImageNative;
-    } else if (m_cliOptions[Option::WindowUnderCursor]) {
-        grabMode = GrabMode::WindowUnderCursor;
-    } else if (Settings::launchAction() == Settings::UseLastUsedCapturemode) {
-        grabMode = toGrabMode(CaptureMode(Settings::captureMode()), transientOnly);
-    }
+    auto cliGrabMode = [&]() -> std::optional<GrabMode> {
+        if (m_cliOptions[Option::Fullscreen]) {
+            return GrabMode::AllScreens;
+        } else if (m_cliOptions[Option::Current]) {
+            return GrabMode::CurrentScreen;
+        } else if (m_cliOptions[Option::ActiveWindow]) {
+            return GrabMode::ActiveWindow;
+        } else if (m_cliOptions[Option::Region]) {
+            return GrabMode::PerScreenImageNative;
+        } else if (m_cliOptions[Option::WindowUnderCursor]) {
+            return GrabMode::WindowUnderCursor;
+        }
+        return std::nullopt;
+    };
+    auto launchActionGrabMode = [&] {
+        switch (Settings::launchAction()) {
+        case Settings::TakeRectangularScreenshot:
+            return GrabMode::PerScreenImageNative;
+        case Settings::TakeFullscreenScreenshot:
+            return GrabMode::AllScreens;
+        case Settings::UseLastUsedCapturemode:
+            return toGrabMode(CaptureMode(Settings::captureMode()), transientOnly);
+        default:
+            return GrabMode::NoGrabModes;
+        }
+    };
+    auto grabMode = cliGrabMode().value_or(m_startMode == StartMode::Background ? GrabMode::AllScreens : launchActionGrabMode());
 
     using RecordingMode = VideoPlatform::RecordingMode;
     RecordingMode recordingMode = RecordingMode::NoRecordingModes;
@@ -859,7 +964,7 @@ void SpectacleCore::takeNewScreenshot(int captureMode, int timeout, bool include
 
 void SpectacleCore::cancelScreenshot()
 {
-    if (m_startMode != StartMode::Gui) {
+    if (m_startMode != StartMode::Gui || !m_returnToViewer) {
         Q_EMIT allDone();
         return;
     }
@@ -869,7 +974,7 @@ void SpectacleCore::cancelScreenshot()
     if (currentTime > 0) {
         SpectacleWindow::setTitleForAll(SpectacleWindow::Previous);
     } else if (!ViewerWindow::instance()) {
-        initViewerWindow(ViewerWindow::Image);
+        initViewerWindow(ViewerWindow::Viewer);
         ViewerWindow::instance()->setVisible(true);
     } else if (ViewerWindow::instance()) {
         Q_EMIT allDone();
@@ -887,10 +992,10 @@ void SpectacleCore::showErrorMessage(const QString &message)
 
 void SpectacleCore::showViewerIfGuiMode(bool minimized)
 {
-    if (m_startMode != StartMode::Gui) {
+    if (m_startMode != StartMode::Gui || !m_returnToViewer) {
         return;
     }
-    initViewerWindow(ViewerWindow::Image);
+    initViewerWindow(ViewerWindow::Viewer);
     if (!m_videoMode && m_cliOptions[CommandLineOptions::EditExisting]) {
         ViewerWindow::instance()->setAnnotating(true);
     }
@@ -985,15 +1090,20 @@ void SpectacleCore::doNotify(ScreenCapture type, const ExportManager::Actions &a
         }
     }
 
-    connect(notification, &QObject::destroyed, this, [this](QObject *notification) {
+    auto onExpired = [this, notification] {
         notifications.removeOne(static_cast<KNotification *>(notification));
         // When there are no more notifications running, we can remove the loop locker.
-        if (notifications.empty()) {
+        if (notifications.empty() && m_eventLoopLocker) {
             QTimer::singleShot(250, this, [this] {
                 m_eventLoopLocker.reset();
             });
         }
-    });
+    };
+    connect(notification, &QObject::destroyed, this, onExpired);
+    // BUG: https://bugs.kde.org/show_bug.cgi?id=503838
+    // We can't call KNotification::close or else the notifications will be removed from history.
+    // 10 seconds is roughly the expected default duration.
+    QTimer::singleShot(10000, notification, onExpired);
 
     notification->sendEvent();
 }
@@ -1144,6 +1254,7 @@ void SpectacleCore::initViewerWindow(ViewerWindow::Mode mode)
 {
     // always switch to gui mode when a viewer window is used.
     m_startMode = SpectacleCore::StartMode::Gui;
+    m_returnToViewer = true;
     deleteWindows();
 
     // Transparency isn't needed for this window.
@@ -1172,6 +1283,12 @@ void SpectacleCore::startRecording(VideoPlatform::RecordingMode mode, bool withP
     if (m_videoPlatform->isRecording() || mode == VideoPlatform::NoRecordingModes) {
         return;
     }
+    if (!CaptureWindow::instances().empty()) {
+        SpectacleWindow::setVisibilityForAll(QWindow::Hidden);
+        if (mode != VideoPlatform::Region) {
+            m_returnToViewer = true;
+        }
+    }
     m_lastRecordingMode = mode;
     setVideoMode(true);
     const auto &output = m_outputUrl.isLocalFile() ? videoOutputUrl() : QUrl();
@@ -1194,7 +1311,12 @@ void SpectacleCore::setVideoMode(bool videoMode)
     if (videoMode == m_videoMode) {
         return;
     }
+    InlineMessageModel::instance()->clear();
     m_videoMode = videoMode;
+    if (!videoMode && m_annotationDocument->baseImage().isNull()) {
+        // Change this if there ends up being a way to toggle video mode outside of rectangle capture mode.
+        takeNewScreenshot(ImagePlatform::PerScreenImageNative, 0, Settings::includePointer(), Settings::includeDecorations(), Settings::includeShadow());
+    }
     Q_EMIT videoModeChanged(videoMode);
 }
 
